#include <fstream>
#include <vector>

#include <cstdlib>
#include <cmath>
#include <sys/time.h>

#include <GL/glut.h>

#include "viewport.h"
#include "camera.h"
#include "vector3.h"
#include "mesh.h"
#include "entity.h"
#include "game.h"
#include "texture.h"


static const int GROUND_DIVISIONS = 20;
static const int WALL_DIVISIONS_X = 64;
static const int WALL_DIVISIONS_Y = 8;
static const float CROSSHAIR_SIZE = 0.001f;


//-- TYPES ---------------------------------------------------------------------
typedef struct material {
    GLfloat ambient[4];
    GLfloat diffuse[4];
    GLfloat specular[4];
    GLfloat shininess[1];
} Material;

typedef struct light {
    float position[4];
    float ambient[4];
    float diffuse[4];
    float specular[4];
    float att_c;    // constant attenuation
    float att_l;    // linear attenuation
    float att_q;    // quadratic attenuation
} Light;


//-- GLOBALS -------------------------------------------------------------------
static Light light0 = {
    {3.0, 5.0, 1.0, 0.0},
    {0.6, 0.6, 0.6, 1.0},
    {0.6, 0.6, 0.6, 1.0},
    {0.8, 0.8, 0.8, 1.0},
    1.0, 0.0, 0.0
};

static Light light1 = {
    {0.8*WORLD_SIZE, 0.25*WALL_HEIGHT, 0.8*WORLD_SIZE, 1.0},
    {0.0, 0.0, 0.0, 0.0},
    {0.6, 0.0, 0.0, 1.0},
    {0.8, 0.0, 0.0, 1.0},
    0.0, 0.3, 0.1
};

static Light light2 = {
    {-0.8*WORLD_SIZE, 0.25*WALL_HEIGHT, 0.8*WORLD_SIZE, 1.0},
    {0.0, 0.0, 0.0, 0.0},
    {0.0, 0.0, 0.6, 1.0},
    {0.0, 0.0, 0.8, 1.0},
    0.0, 0.3, 0.1
};

static Light light3 = {
    {0.8*WORLD_SIZE, 0.25*WALL_HEIGHT, -0.8*WORLD_SIZE, 1.0},
    {0.0, 0.0, 0.0, 0.0},
    {0.0, 0.6, 0.0, 1.0},
    {0.0, 0.8, 0.0, 1.0},
    0.0, 0.3, 0.1
};

static Light light4 = {
    {-0.8*WORLD_SIZE, 0.25*WALL_HEIGHT, -0.8*WORLD_SIZE, 1.0},
    {0.0, 0.0, 0.0, 0.0},
    {0.6, 0.0, 0.6, 1.0},
    {0.8, 0.0, 0.8, 1.0},
    0.0, 0.3, 0.1
};

static GLuint dl_bunny, dl_world;
Image texture;


//-- PROTOTYPES ----------------------------------------------------------------
static void setMaterials(const Material *mat);
static void setupLight(int id, const Light *light);
static void initDisplayLists(void);

static void drawWorld(void);
static void drawCrosshairs(void);
static void drawBunnies(const std::vector<Entity *> &bunnies, float size);
static void drawBullets(void);
static void display(void);
static void reshape(int w, int h);

static void mouse(int button, int state, int x, int y);
static void mouseMove(int x, int y);
static void keyboardDown(unsigned char key, int x, int y);
static void keyboardUp(unsigned char key, int x, int y);


//-- DISPLAY -------------------------------------------------------------------
static void setMaterials(const Material *mat)
{
    glMaterialfv(GL_FRONT, GL_AMBIENT, mat->ambient);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, mat->diffuse);
    glMaterialfv(GL_FRONT, GL_SPECULAR, mat->specular);
    glMaterialfv(GL_FRONT, GL_SHININESS, mat->shininess);
}

static void setupLight(int id, const Light *light)
{
    glEnable(id);
    glLightfv(id, GL_DIFFUSE, light->diffuse);
    glLightfv(id, GL_AMBIENT, light->ambient);
    glLightfv(id, GL_SPECULAR, light->specular);
    glLightf(id, GL_CONSTANT_ATTENUATION, light->att_c);
    glLightf(id, GL_LINEAR_ATTENUATION, light->att_l);
    glLightf(id, GL_QUADRATIC_ATTENUATION, light->att_q);
}

static void initDisplayLists(void)
{
    dl_bunny = glGenLists(2);
    dl_world = dl_bunny + 1;

    glNewList(dl_bunny, GL_COMPILE);
    {
        /* raise to ground plane */
        glTranslatef(0.0, 0.35, 0.0);

        /* scale to unit-size */
        glScalef(1.0/bunnyMesh->max_extent, 1.0/bunnyMesh->max_extent, 1.0/bunnyMesh->max_extent);

        /* center to origin */
        glTranslatef(-bunnyMesh->center.x(), -bunnyMesh->center.y(), -bunnyMesh->center.z());
        bunnyMesh->draw();
    }
    glEndList();

    glNewList(dl_world, GL_COMPILE);
    {
        drawWorld();
    }
    glEndList();
}

static void drawWorld(void)
{
    static Material mat_ground = {
        {0.2, 0.2, 0.2, 1.0},
        {0.5, 0.5, 0.5, 0.0},
        {0.3, 0.3, 0.3, 1.0},
        {1.0}
    };
    static Material mat_wall = {
        {0.2, 0.2, 0.2, 1.0},
        {1.0, 1.0, 1.0, 1.0},
        {0.0, 0.0, 0.0, 0.0},
        {0.0}
    };

    /* ground plane */
    setMaterials(&mat_ground);
    glPushMatrix();
    {
        glBindTexture(GL_TEXTURE_2D, 0);
        glBegin(GL_QUADS);
        {
            for (int i = 0; i < GROUND_DIVISIONS; i++) {
                for (int j = 0; j < GROUND_DIVISIONS; j++) {
                    float xStride = 2.0 * WORLD_SIZE / (double)GROUND_DIVISIONS;
                    float yStride = 2.0 * WORLD_SIZE / (double)GROUND_DIVISIONS;
                    float xStart = -WORLD_SIZE + i * xStride;
                    float yStart = -WORLD_SIZE + j * yStride;

                    glNormal3f(0.0, 1.0, 0.0);
                    glTexCoord2f(0.0, 0.0);
                    glVertex3f(xStart + 0 * xStride, 0.0, yStart + 0 * yStride);

                    glNormal3f(0.0, 1.0, 0.0);
                    glTexCoord2f(1.0, 0.0);
                    glVertex3f(xStart + 1 * xStride, 0.0, yStart + 0 * yStride);

                    glNormal3f(0.0, 1.0, 0.0);
                    glTexCoord2f(1.0, 1.0);
                    glVertex3f(xStart + 1 * xStride, 0.0, yStart + 1 * yStride);

                    glNormal3f(0.0, 1.0, 0.0);
                    glTexCoord2f(0.0, 1.0);
                    glVertex3f(xStart + 0 * xStride, 0.0, yStart + 1 * yStride);
                }
            }
        }
        glEnd();
    }
    glPopMatrix();


    /* east wall */
    setMaterials(&mat_wall);
    glBindTexture(GL_TEXTURE_2D, 1);
    glPushMatrix();
    {
        glTranslatef(+WORLD_SIZE, 0.0, 0.0);

        glBegin(GL_QUADS);
        {
            for (int i = 0; i < WALL_DIVISIONS_X; i++) {
                for (int j = 0; j < WALL_DIVISIONS_Y; j++) {
                    float xStride = 2.0 * WORLD_SIZE / (double)WALL_DIVISIONS_X;
                    float yStride = WALL_HEIGHT / (double)WALL_DIVISIONS_Y;
                    float xStart = -WORLD_SIZE + i * xStride;
                    float yStart = 0.0 + j * yStride;
            
                    glNormal3f(-1.0, 0.0, 0.0);
                    glTexCoord2f(0.0, 0.0);
                    glVertex3f(0.0, yStart + 0 * yStride, xStart + 0 * xStride);

                    glNormal3f(-1.0, 0.0, 0.0);
                    glTexCoord2f(0.0, 1.0);
                    glVertex3f(0.0, yStart + 1 * yStride, xStart + 0 * xStride);

                    glNormal3f(-1.0, 0.0, 0.0);
                    glTexCoord2f(1.0, 1.0);
                    glVertex3f(0.0, yStart + 1 * yStride, xStart + 1 * xStride);

                    glNormal3f(-1.0, 0.0, 0.0);
                    glTexCoord2f(1.0, 0.0);
                    glVertex3f(0.0, yStart + 0 * yStride, xStart + 1 * xStride);
                }
                
            }
        }
        glEnd();
    }
    glPopMatrix();

    /* west wall */
    glPushMatrix();
    {
        glTranslatef(-WORLD_SIZE, 0.0, 0.0);

        glBegin(GL_QUADS);
        {
            for (int i = 0; i < WALL_DIVISIONS_X; i++) {
                for (int j = 0; j < WALL_DIVISIONS_Y; j++) {
                    float xStride = 2.0 * WORLD_SIZE / (double)WALL_DIVISIONS_X;
                    float yStride = WALL_HEIGHT / (double)WALL_DIVISIONS_Y;
                    float xStart = -WORLD_SIZE + i * xStride;
                    float yStart = 0.0 + j * yStride;
            
                    glNormal3f(+1.0, 0.0, 0.0);
                    glTexCoord2f(0.0, 0.0);
                    glVertex3f(0.0, yStart + 0 * yStride, xStart + 1 * xStride);

                    glNormal3f(+1.0, 0.0, 0.0);
                    glTexCoord2f(0.0, 1.0);
                    glVertex3f(0.0, yStart + 1 * yStride, xStart + 1 * xStride);

                    glNormal3f(+1.0, 0.0, 0.0);
                    glTexCoord2f(1.0, 1.0);
                    glVertex3f(0.0, yStart + 1 * yStride, xStart + 0 * xStride);

                    glNormal3f(+1.0, 0.0, 0.0);
                    glTexCoord2f(1.0, 0.0);
                    glVertex3f(0.0, yStart + 0 * yStride, xStart + 0 * xStride);
                }
                
            }
        }
        glEnd();
    }
    glPopMatrix();

    /* north wall */
    glPushMatrix();
    {
        glTranslatef(0.0, 0.0, +WORLD_SIZE);

        glBegin(GL_QUADS);
        {
            for (int i = 0; i < WALL_DIVISIONS_X; i++) {
                for (int j = 0; j < WALL_DIVISIONS_Y; j++) {
                    float xStride = 2.0 * WORLD_SIZE / (double)WALL_DIVISIONS_X;
                    float yStride = WALL_HEIGHT / (double)WALL_DIVISIONS_Y;
                    float xStart = -WORLD_SIZE + i * xStride;
                    float yStart = 0.0 + j * yStride;
            
                    glNormal3f(0.0, 0.0, -1.0);
                    glTexCoord2f(0.0, 0.0);
                    glVertex3f(xStart + 1 * xStride, yStart + 0 * yStride, 0.0);

                    glNormal3f(0.0, 0.0, -1.0);
                    glTexCoord2f(0.0, 1.0);
                    glVertex3f(xStart + 1 * xStride, yStart + 1 * yStride, 0.0);

                    glNormal3f(0.0, 0.0, -1.0);
                    glTexCoord2f(1.0, 1.0);
                    glVertex3f(xStart + 0 * xStride, yStart + 1 * yStride, 0.0);

                    glNormal3f(0.0, 0.0, -1.0);
                    glTexCoord2f(1.0, 0.0);
                    glVertex3f(xStart + 0 * xStride, yStart + 0 * yStride, 0.0);
                }
                
            }
        }
        glEnd();
    }
    glPopMatrix();

    /* south wall */
    glPushMatrix();
    {
        glTranslatef(0.0, 0.0, -WORLD_SIZE);

        glBegin(GL_QUADS);
        {
            for (int i = 0; i < WALL_DIVISIONS_X; i++) {
                for (int j = 0; j < WALL_DIVISIONS_Y; j++) {
                    float xStride = 2.0 * WORLD_SIZE / (double)WALL_DIVISIONS_X;
                    float yStride = WALL_HEIGHT / (double)WALL_DIVISIONS_Y;
                    float xStart = -WORLD_SIZE + i * xStride;
                    float yStart = 0.0 + j * yStride;
            
                    glNormal3f(0.0, 0.0, +1.0);
                    glTexCoord2f(0.0, 0.0);
                    glVertex3f(xStart + 0 * xStride, yStart + 0 * yStride, 0.0);

                    glNormal3f(0.0, 0.0, +1.0);
                    glTexCoord2f(0.0, 1.0);
                    glVertex3f(xStart + 0 * xStride, yStart + 1 * yStride, 0.0);

                    glNormal3f(0.0, 0.0, +1.0);
                    glTexCoord2f(1.0, 1.0);
                    glVertex3f(xStart + 1 * xStride, yStart + 1 * yStride, 0.0);

                    glNormal3f(0.0, 0.0, +1.0);
                    glTexCoord2f(1.0, 0.0);
                    glVertex3f(xStart + 1 * xStride, yStart + 0 * yStride, 0.0);
                }
                
            }
        }
        glEnd();
    }
    glPopMatrix();
}

static void drawCrosshairs(void)
{
    glPushMatrix();
    {
        // ignore camera transform
        glLoadIdentity();
        glDisable(GL_LIGHTING);
        glBegin(GL_LINES);
        {
            glColor3f(1.0, 0.0, 0.0);
            glVertex3f(-CROSSHAIR_SIZE, 0.0, -0.1);
            glVertex3f(+CROSSHAIR_SIZE, 0.0, -0.1);

            glVertex3f(0.0, -CROSSHAIR_SIZE, -0.1);
            glVertex3f(0.0, +CROSSHAIR_SIZE, -0.1);
        }
        glEnd();
        glEnable(GL_LIGHTING);
    }
    glPopMatrix();
}

static void drawBunnies(const std::vector<Entity *> &bunnies, float size)
{
    static Material mat_bunny = {
        {0.3, 0.3, 0.3, 1.0},
        {1.0, 1.0, 1.0, 1.0},
        {1.0, 1.0, 1.0, 1.0},
        {2.0}
    };


    setMaterials(&mat_bunny);
    glPushMatrix();
    {
        std::vector<Entity *>::const_iterator itr;
        for (itr = bunnies.begin(); itr != bunnies.end(); itr++) {
            glPushMatrix();
            {
                (*itr)->doTransforms();
                glScalef(size, size, size);
                glCallList(dl_bunny);
            }
            glPopMatrix();
        }
    }
    glPopMatrix();
}

static void drawBullets(void)
{
    static Material mat_bullet = {
        {0.081176, 0.208235, 0.235294, 1.0},
        {0.081176, 0.208235, 0.235294, 1.0},
        {0.270588, 0.694118, 0.784314, 1.0},
        {10.0}
    };

    setMaterials(&mat_bullet);

    std::vector<Movable *>::iterator itr;
    for (itr = bullets.begin(); itr != bullets.end(); itr++) {
        glPushMatrix();
        {
            (*itr)->doTransforms();

            glutSolidSphere(BULLET_SIZE, 16, 16);
        }
        glPopMatrix();
    }
}

static void display(void)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();
    {
        // camera transform
        Vector3 eye = camera.eye();
        Vector3 lookAt = camera.lookAt();
        Vector3 up = camera.up();

        gluLookAt(eye.x(), eye.y(), eye.z(),
                lookAt.x(), lookAt.y(), lookAt.z(),
                up.x(), up.y(), up.z());

        // set lighting
        glLightfv(GL_LIGHT0, GL_POSITION, light0.position);
        glLightfv(GL_LIGHT1, GL_POSITION, light1.position);
        glLightfv(GL_LIGHT2, GL_POSITION, light2.position);
        glLightfv(GL_LIGHT3, GL_POSITION, light3.position);
        glLightfv(GL_LIGHT4, GL_POSITION, light4.position);

        // draw world
        glPushMatrix();
        {
            glCallList(dl_world);
        }
        glPopMatrix();

        // draw bunnies_big
        glPushMatrix();
        {
            drawBunnies(bunnies_big, 2.0 * BUNNY_SIZE_BIG);
            drawBunnies(bunnies_medium, 2.0 * BUNNY_SIZE_MEDIUM);
            drawBunnies(bunnies_small, 2.0 * BUNNY_SIZE_SMALL);
        }
        glPopMatrix();

        // draw bullets
        glPushMatrix();
        {
            drawBullets();
        }
        glPopMatrix();
    }
    glPopMatrix();

    drawCrosshairs();
    glutSwapBuffers();
}

static void reshape(int w, int h)
{
    view.set(w, h);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(60.0, view.ratio(), view.near(), view.far());

    glMatrixMode(GL_MODELVIEW);
    glViewport(0, 0, w, h);
    glutPostRedisplay();
}


//-- INPUT ---------------------------------------------------------------------
static void mouse(int button, int state, int x, int y)
{
    if (button == GLUT_LEFT_BUTTON) {
        if (state == GLUT_DOWN) {
            Vector3 position = camera.eye() + camera.view().unit() * 2.0 * PLAYER_SIZE;
            Vector3 velocity = camera.view().unit() * BULLET_SPEED;
            bullets.push_back(new Movable(position, velocity));
        }
    }
}

static void mouseMove(int x, int y)
{
    double yaw = (x - view.width()/2) * 90.0 / 800.0;
    double pitch = -(y - view.height()/2) * 90.0 / 800.0;

    camera.yaw(yaw);
    camera.pitch(pitch);
    glutPostRedisplay();
}

static void keyboardDown(unsigned char key, int x, int y)
{
    Vector3 v;

    switch (key) {
    case 'q': case 'Q':
        exit(EXIT_SUCCESS);
        break;

    case 'w': case 'W':
        moveForward = 1;
        break;

    case 's': case 'S':
        moveBackward = 1;
        break;

    case 'a': case 'A':
        strafeLeft = 1;
        break;

    case 'd': case 'D':
        strafeRight = 1;
        break;
    }
}

static void keyboardUp(unsigned char key, int x, int y)
{
    switch (key) {
    case 'w': case 'W':
        moveForward = 0;
        break;

    case 's': case 'S':
        moveBackward = 0;
        break;

    case 'a': case 'A':
        strafeLeft = 0;
        break;

    case 'd': case 'D':
        strafeRight = 0;
        break;
    }
}


//-- MAIN ----------------------------------------------------------------------
int main(int argc, char **argv)
{
    // set up my window
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
    glutInitWindowSize(view.width(), view.height());
    glutInitWindowPosition(100, 100);
    glutCreateWindow("Bunnyroids");
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glutWarpPointer(view.width()/2, view.width()/2);

    // register glut callback functions
    glutDisplayFunc(display);
    glutReshapeFunc(reshape);
    glutKeyboardFunc(keyboardDown);
    glutKeyboardUpFunc(keyboardUp);
    glutWarpPointer(view.width()/2, view.height()/2);
    glutSetCursor(GLUT_CURSOR_NONE);
    glutMouseFunc(mouse);
    glutMotionFunc(mouseMove);
    glutPassiveMotionFunc(mouseMove);
    glutIdleFunc(physics);
    glutIgnoreKeyRepeat(1);
    glutFullScreen();
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_2D);

    // initialize lighting
    glEnable(GL_LIGHTING);
    setupLight(GL_LIGHT0, &light0);
    setupLight(GL_LIGHT1, &light1);
    setupLight(GL_LIGHT2, &light2);
    setupLight(GL_LIGHT3, &light3);
    setupLight(GL_LIGHT4, &light4);

    glShadeModel(GL_SMOOTH);
    glEnable(GL_NORMALIZE);

    /* load textures */
    LoadTexture(&texture, "grass-texture.bmp", 0);
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, texture.sizeX, texture.sizeY,
            GL_RGB, GL_UNSIGNED_BYTE, texture.data);
    delete texture.data;

    LoadTexture(&texture, "cinder-block-texture.bmp", 1);
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, texture.sizeX, texture.sizeY,
            GL_RGB, GL_UNSIGNED_BYTE, texture.data);
    delete texture.data;

    LoadTexture(&texture, "fur-texture.bmp", 2);
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, texture.sizeX, texture.sizeY,
            GL_RGB, GL_UNSIGNED_BYTE, texture.data);
    delete texture.data;

    /* read bunny mesh */
    std::ifstream fin("bunny500.m");
    bunnyMesh = Mesh::read(fin);
    initDisplayLists();

    /* create bunnies */
    for (int i = 0; i < 7; i++) {
        /* position */
        double r_min = PLAYER_SIZE + BUNNY_SIZE_BIG + 3.0 * BUNNY_SPEED_BIG;
        double r_max = WORLD_SIZE - BUNNY_SIZE_BIG;
        double t = 2.0 * M_PI * (double)rand() / RAND_MAX;
        double r = r_min + (double)rand() / RAND_MAX * (r_max - r_min);
        double x = r * cos(t);
        double z = r * sin(t);

        /* orientation */
        double o = 360.0 * (double)rand() / RAND_MAX;
        
        Entity *e = new Entity(bunnyMesh, x, 0.0, z);
        e->rotate(o, Vector3(0.0, 1.0, 0.0));
        Vector3 velocity = e->forward().unit() * BUNNY_SPEED_BIG;
        e->setVelocity(velocity.x(), velocity.y(), velocity.z());
        bunnies_big.push_back(e);
    }

    glutMainLoop();
}
