#include "mesh.h"

#include <iostream>
#include <vector>

#include <cstdlib>
#include <cstring>
#include <cassert>


static void readLine(Mesh *mesh, char *str);


void Mesh::draw(void) const
{
    for (unsigned int i = 0; i < faces.size(); i++) {
        drawTriangle(faces[i]);
    }
}

void Mesh::drawNormals(void) const
{
    for (unsigned int i = 0; i < faces.size(); i++) {
        drawNormal(faces[i]);
    }
}

void Mesh::drawTriangle(const Triangle &t) const
{
    Vector3 n;

    glBindTexture(GL_TEXTURE_2D, 2);
    glBegin(GL_TRIANGLES);
    {
        n = normals[t.v1 - 1];
        glNormal3f(n.x(), n.y(), n.z());
        glTexCoord2f(0.0, 0.0);
        glVertex3f(vertices[t.v1 - 1].x(),
            vertices[t.v1 - 1].y(),
            vertices[t.v1 - 1].z());

        n = normals[t.v2 - 1];
        glNormal3f(n.x(), n.y(), n.z());
        glTexCoord2f(0.0, 1.0);
        glVertex3f(vertices[t.v2 - 1].x(),
            vertices[t.v2 - 1].y(),
            vertices[t.v2 - 1].z());

        n = normals[t.v3 - 1];
        glNormal3f(n.x(), n.y(), n.z());
        glTexCoord2f(1.0, 0.5);
        glVertex3f(vertices[t.v3 - 1].x(),
            vertices[t.v3 - 1].y(),
            vertices[t.v3 - 1].z());
    }
    glEnd();
}

void Mesh::drawNormal(const Triangle &t) const
{
    Vector3 v1 = vertices[t.v1 - 1];
    Vector3 v2 = vertices[t.v2 - 1];
    Vector3 v3 = vertices[t.v3 - 1];
    Vector3 center = (v1 + v2 + v3) / 3.0;
    Vector3 normTip = center + t.normal * max_extent*max_extent / 10.0;

    glBegin(GL_LINES);
        glColor3f(1.0, 0.0, 0.0);
        glVertex3f(center.x(), center.y(), center.z());
        glVertex3f(normTip.x(), normTip.y(), normTip.z());
    glEnd();
}

Mesh *Mesh::read(istream &is)
{
    Mesh *mesh = new Mesh;
    char str[256];

    for (;is;) {
        is >> std::ws;
        is.get(str, sizeof(str));
        if (!is)
            break;
        is.ignore(9999, '\n');
        readLine(mesh, str);
    }

    /* calculate max extent */
    mesh->max_extent = mesh->max_x - mesh->min_x;
    if (mesh->max_y - mesh->min_y > mesh->max_extent)
        mesh->max_extent = mesh->max_y - mesh->min_y;

    /* finish calculating center of mass */
    mesh->center /= mesh->vertices.size();

    /* normalize vertex normal vectors */
    for (unsigned int i = 0; i < mesh->normals.size(); i++) {
        mesh->normals[i].normalize();
    }

    /* scale normal vectors by max_extent */
    for (unsigned int i = 0; i < mesh->faces.size(); i++) {
        mesh->faces[i].normal /= mesh->max_extent;
    }
    for (unsigned int i = 0; i < mesh->normals.size(); i++) {
        mesh->normals[i] /= mesh->max_extent;
    }

    return mesh;
}

static void readLine(Mesh *mesh, char* str)
{
    int vi;
    float x, y, z;
    float r, g, b;
    int mat;

    if (str[0]=='#')
        return;

    //read a vertex or face
    if (str[0]=='V' && !strncmp(str,"Vertex ",7)) {
        assert(sscanf(str, "Vertex %d %g %g %g", &vi, &x, &y, &z) == 4);
        Vector3 v(x, y, z);

        //store the vertex
        mesh->addVertex(v);

        //house keeping to display in center of the scene
        mesh->center += v;

        if (v.x() > mesh->max_x)
            mesh->max_x = v.x();
        if (v.x() < mesh->min_x)
            mesh->min_x = v.x();
        if (v.y() > mesh->max_y)
            mesh->max_y = v.y();
        if (v.y() < mesh->min_y)
            mesh->min_y = v.y();
        if (v.z() > mesh->max_z)
            mesh->max_z = v.z();
        if (v.z() < mesh->min_z)
            mesh->min_z = v.z();
    }
    else if (str[0]=='F' && !strncmp(str,"Face ",5)) {
        Triangle t;
        char *s = str+4;
        int fi = -1;

        for (int t_i = 0; ; t_i++) {
            while (*s && isspace(*s))
                s++;

            // if we reach the end of the line break out of the loop
            if (!*s)
                break;

            // save the position of the current character
            char *beg = s;

            // advance to next space
            while (*s && isdigit(*s))
                s++;

            // covert the character to an integer
            int j = atoi(beg);

            // the first number we encounter will be the face index, don't store it
            if (fi < 0) {
                fi = j;
                continue;
            }

            // otherwise process the digit we've grabbed in j as a vertex index
            // the first number will be the face id the following are vertex ids
            if (t_i == 1)
                t.v1 = j;
            else if (t_i == 2)
                t.v2 = j;
            else if (t_i == 3)
                t.v3 = j;

            // if there is more data to process break out
            if (*s =='{')
                break;
        }

        // possibly process colors if the mesh has colors
        if (*s && *s == '{') {
            char *s1 = s+1;
            std::cout << "trying to parse color " << !strncmp(s1, "rgb", 3) << std::endl;

            // if we're reading off a color
            if (!strncmp(s1, "rgb=", 4)) {
                // grab the values of the string
                assert(sscanf(s1, "rgb=(%g %g %g) matid=%d", &r, &g, &b, &mat) == 4);
                t.color.set(r, g, b);
                std::cout << "set color to: " << r << " " << g << " " << b << std::endl;
            }
        }

        // calculate normal vector
        Vector3 u = mesh->vertices[t.v1 - 1];
        Vector3 v = mesh->vertices[t.v2 - 1];
        Vector3 w = mesh->vertices[t.v3 - 1];
        Vector3 n = (v - u).cross(w - u);
        t.normal = n.unit();

        // add face normal to vertex normals
        mesh->normals[t.v1 - 1] += t.normal;
        mesh->normals[t.v2 - 1] += t.normal;
        mesh->normals[t.v3 - 1] += t.normal;

        // store the triangle read in
        mesh->addTriangle(t);
    }
}
